/*
   This file is part of the index package,
   Copyright (C) 1997 Harold Tay, haroldtay@pacific.net.sg

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA
 */
/* Given one or more words on the command line, search for those
   words in the index file on fd 0.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <regex.h>
#include "index.h"
#include "hash.h"
#include "codec.h"

char * map;
int mapsize;
char * locations;
char * words;
char * end;
int verbose = 1;
int printfrom = 0;
int printto = 999999;

enum{ SUBSTRING=0, WHOLEWORD, REGEX, PREFIX };

void
findem(int nr_patterns, char ** patterns, int matchtype, regex_t * re)
{
  int i;
  char * word = words;
  int hds[8];
  unsigned int hash(char * text){ return((unsigned)text); }
  int cmp(const char * a, const char * b){
    return((unsigned)a - (unsigned)b); }

  for(i = 0; i < nr_patterns; i++)
    hds[i] = hash_create(hash, cmp, 256);

  while( word < end ){
    int * lengths = NULL;
    int wlen;

    if( PREFIX == matchtype ){
      lengths = malloc(nr_patterns * sizeof(int));
      for(i = 0; i < nr_patterns; i++)
        lengths[i] = strlen(patterns[i]);
    }

    wlen = strlen(word);

    for(i = 0; i < nr_patterns; i++){
      int match;
      switch( matchtype ){
      case WHOLEWORD: match = (0 == strcmp(word, patterns[i])); break;
      case REGEX: match = (0 == regexec(&re[i], word, 0, NULL, 0)); break;
      case PREFIX:
        match = (0 == strncmp(word, patterns[i], lengths[i])); break;
      default:  /* fall through */
      case SUBSTRING: match = (strstr(word, patterns[i]) != NULL); break;
      };

      if( match ){
        int * ary;
        int arlen;
        int j;

        codec_dec(word + 1 + wlen, &ary, &arlen);

        /* if a location occurs more than once per pattern,
           record only the first */
        for(j = 0; j < arlen; j++){
          int c;
          c = (int)hash_fetch(hds[i], (char *)ary[j]);
          if( -1 == c )hash_insert(hds[i], (char *)ary[j], (char *)1);
          /*
          if( -1 != c )hash_replace(hds[i], (char *)ary[j], (char *)(c + 1));
          else hash_insert(hds[i], (char *)ary[j], (char *)1);
           */
        }
      }
    }
    word += (wlen + 1);
    word += codec_skip(word);
  }

  {
    int merge(char * k, char * v){
      int c = (int)hash_fetch(hds[0], k);
      if( -1 != c )hash_replace(hds[0], k, (char *)(c + 1));
      else hash_insert(hds[0], k, v);
      return(0);
    }

    if( verbose ){
      for(i = 0; i < nr_patterns; i++)
        printf("%s %d\n", patterns[i], hash_size(hds[i]));
      printf("\n");
    }
    for(i = 1; i < nr_patterns; i++){
      hash_walk(hds[i], merge);
      hash_destroy(hds[i]);
    }
  }
  {
    struct m{
      int nr;
      unsigned loc;
    } * m;
    int nr = 0;
    int makearray(char * k, char * v){
      m[nr].loc = (unsigned)k; m[nr].nr = (int)v; nr++; return(0); }
    int numerically(const void * a, const void * b){
      return(((struct m *)b)->nr - ((struct m *)a)->nr); }

    m = malloc(hash_size(hds[0]) * sizeof(struct m));
    hash_walk(hds[0], makearray);
    hash_destroy(hds[0]);

    if( verbose )printf("%d\n", nr);

    qsort(m, nr, sizeof(struct m), numerically);
    for(i = printfrom; i < printto && i < nr; i++)
      printf("%d %s\n", m[i].nr, locations + m[i].loc);
    free(m);
  }
}

static void
cleanup(void)
{
  if( map ){ munmap(map, mapsize); map = NULL; }
  locations = words = end = NULL;
}

static void
init(void)
{
  struct stat st;
  struct meta meta;

  if( -1 == fstat(0, &st) ){
    fprintf(stderr, "Can't stat 0: %s\n", strerror(errno));
    exit(1);
  }
  mapsize = st.st_size;
  errno = 0;
  map = mmap(0, st.st_size, PROT_READ, MAP_FILE|MAP_PRIVATE, 0, 0);
  if( (int)map <= 0 ){
    fprintf(stderr, "Can't map 0: %s\n", strerror(errno));
    exit(1);
  }
  memcpy(&meta, map, sizeof(meta));
  if( strcmp(meta.magic, MAGIC) ){
    fprintf(stderr, "Bad magic in file\n");
    munmap(map, st.st_size);
    exit(2);
  }
  locations = map + sizeof(meta);
  words = locations + meta.locations_size;
  end = words + meta.words_size;

}

char * progname;

static void
usage(void)
{
  fprintf(stderr,
    "Usage: %s [-q] [-n-m] [-f file] [-rw] word ...\n"
    "where n and m are range numbers\n", progname);
  exit(1);
}

int
main(int argc, char ** argv)
{
  int i;
  int matchtype = SUBSTRING;
  char * file = NULL;
  regex_t * re = NULL;
  char ** patterns;
  int nr_patterns;
  char * cp;

  progname = ((cp = strrchr(argv[0], '/')) ? cp + 1 : argv[0]);

  for(i = 1; i < argc; i++)
    if( '-' == *argv[i] ){
      switch( argv[i][1] ){
      case 'p': matchtype = PREFIX; break;
      case 'r': matchtype = REGEX; break;
      case 'w': matchtype = WHOLEWORD; break;
      case 'q': verbose = 0; break;
      case 'f': file = argv[++i]; continue;
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
        printfrom = atoi(&argv[i][1]);
        cp = strchr(&argv[i][1], '-');
        if( cp )printto = atoi(cp + 1);
        continue;
      default:
        usage();
      }
      if( argv[i][2] )usage();
    }else
      break;

  if( file ){
    int fd = open(file, O_RDONLY);
    if( -1 == fd ){
      fprintf(stderr, "Can't open %s: %s\n", file, strerror(errno));
      exit(1);
    }
    if( -1 == dup2(fd, 0) ){
      fprintf(stderr, "Can't dup: %s\n", strerror(errno));
      exit(1);
    }
  }

  nr_patterns = argc - i;
  if( nr_patterns <= 0 ){
    fprintf(stderr, "Need at least one pattern to search for\n");
    exit(1);
  }else if( nr_patterns > 8 ){
    fprintf(stderr, "Max 8 patterns please\n");
    exit(1);
  }
  patterns = argv + i;
  for(i = 0; i < nr_patterns; i++){
    char * cp;
    if( ! patterns[i][0] || ! patterns[i][1] ){
      fprintf(stderr, "Patterns should be at least 2 characters long\n");
      cleanup(); exit(1);
    }
    for(cp = patterns[i]; *cp; cp++)
      if( isupper(*cp) )*cp = tolower(*cp);
  }

  if( REGEX == matchtype ){
    re = malloc(nr_patterns * sizeof(regex_t));
    for(i = 0; i < nr_patterns; i++)
      if( regcomp(&re[i], patterns[i], REG_EXTENDED|REG_NOSUB) ){
        fprintf(stderr, "Bad regex \"%s\"\n", patterns[i]);
        free(re);
        exit(1);
      }
  }

  init();

  findem(nr_patterns, patterns, matchtype, re);

  if( re ){
    for(i = 0; i < nr_patterns; i++)regfree(&re[i]); free(re); }
  cleanup();
  exit(0);
}

